# json-indent

[![EditorConfig-enabled](https://img.shields.io/badge/EditorConfig-enabled-brightgreen?logo=EditorConfig&logoColor=white)](https://editorconfig.org/)
[![pre-commit-enabled](https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white)](https://github.com/pre-commit/pre-commit)
[![Python-3.7+](https://img.shields.io/badge/Python-3.7+-informational?logo=Python&logoColor=white)](https://www.python.org)
[![CodeStyle-ruff](https://img.shields.io/badge/CodeStyle-ruff-informational)](https://astral.sh/ruff)
[![pre-commit-hook-included](https://img.shields.io/badge/pre--commit--hook-included-brightgreen?logo=pre-commit&logoColor=white)](#pre-commit-hook)


This is a simple Python-based tool for parsing/indenting/formatting JSON.

Python's `json` module actually contains such a tool; it's even simpler and is
missing twiddles for specifying the amount of indent, whether to (re)compact,
and whether to sort keys when parsing.

**json-indent** can be useful:

- On the command-line
- From within [Vim][] or similar editors
- As a [pre-commit][] hook

([js-beautify][] is another such tool; it can parse JavaScript as well as
JSON, but it has no built-in [pre-commit][] hook).


[begintoc]: #

## Contents

- [Requirements](#requirements)
- [Installing](#installing)
- [How to Use](#how-to-use)
    - [Quickstart](#quickstart)
- [Command-line Autocompletion](#command-line-autocompletion)
- [Advanced Topics](#advanced-topics)
    - [Pre-Commit Hook](#pre-commit-hook)
    - [Integration with Vim](#integration-with-vim)
- [Developing json-indent](#developing-json-indent)
- [References](#references)

[endtoc]: # (Generated by markdown-toc pre-commit hook)


## Requirements

- A [Python] interpreter, version 3.7 or later (version 2.7 support is
  deprecated, but may still work)


## Installing

It is recommended to install **json-indent** either:

- In its own virtual environment (for example, using [uv][]), or
- As a [pre-commit hook](#pre-commit-hook)

To install with `uv`:

    uv tool install json-indent


## How to Use

For a summary of all the available options:

    uv tool run json-indent --help

Or, with the `uvx` shortcut:

    uvx json-indent --help

Or, if you prefer to call the script directly:

    /path/to/json-indent --help

> [!TIP]
>
> There are a number of different ways to run tools that are delivered as
> Python modules.  For the remainder of this document, we will use
> `uvx json-indent` to mean whichever one you prefer for your Python
> installation and operating environment.
>
> See [How to Run Your Python Scripts][run-python-scripts] for more details.


### Quickstart

To read JSON from a file, parse it, and write it to the standard output:

    uvx json-indent input.json

To write to a file instead:

    uvx json-indent -o output.json input.json

To read from the standard input, you can use `-` as the filename:

    uvx json-indent -o - -

Or simply:

    uvx json-indent

To modify a file in place:

    uvx json-indent --inplace input.json

To display `json-indent`'s version:

    uvx json-indent --version

See the command-line help for more options, including:

- Indent size
- "Compact" mode
- Key sorting


## Command-line Autocompletion

**json-indent** provides command-line [autocompletion][] for the [Bash][]
command-line shell and compatible shells, using the
[argcomplete][argcomplete-pypi] Python package.

For instructions on how to enable completion:

    uvx json-indent --completion-help


## Advanced Topics

### Pre-Commit Hook

**json-indent** has support for the [pre-commit][] framework for running tools
before committing to your git project.  The hook will:

1. Parse your JSON files and emit messages about any errors it encounters.
2. Format and indent your JSON files according to the `--indent`, `--compact`,
    and `--sort-keys` options you supply (or their default values if you don't
    supply one).

To add a hook for **json-indent** to your project, use the following YAML in
your project's [.pre-commit-config.yaml](examples/.pre-commit-config.yaml):

```yaml
  - repo: https://github.com/jmknoble/json-indent
    rev: v2.4.0
    hooks:
      - id: json-indent
```

> [!NOTE]
>
> **HOW IT WORKS:**
>
> Under the hood this uses a [.pre-commit-hooks.yaml](.pre-commit-hooks.yaml)
> file in the root of the Git project to define the parameters of the
> pre-commit hook.
>
> Key info:
>
> - **json-indent** must accept multiple input files on the command line.
> - The `--inplace` option is needed in order to automatically indent JSON
>   source files.


### Integration with Vim

You can use **json-indent** from within [Vim][] to format JSON directly in
your editing session.

To enable this functionality, you can add the following to your [vimrc][]
file:

```viml
""""""""""""""""""""""""""""""""""""""""
" JSON parsing/indenting
" https://github.com/jmknoble/json-indent
"
" Normal mode
"
" Compact JSON stanza starting at current character (uses '%' to move to matching brace)
nmap <Leader>jc :%!json-indent -c
nmap <Leader>JC <Leader>jc
"
" Parse/indent JSON stanza contained in current line
nmap <Leader>jp 0!$json-indent<CR>j
nmap <Leader>JP <Leader>jp
"
" Parse/indent JSON stanza contained in current line with sorted keys
nmap <Leader>js 0!$json-indent -s<CR>j
nmap <Leader>JS <Leader>js
"
" Visual mode
"
" Compact JSON stanza contained in current selection
vmap <Leader>jc !json-indent -c<CR>j
vmap <Leader>JC <Leader>jc
"
" Parse/indent JSON stanza contained in current selection
vmap <Leader>jp !json-indent
vmap <Leader>JP <Leader>jp
"
" Parse/indent JSON stanza contained in current selection with sorted keys
vmap <Leader>js !json-indent -s<CR>j
vmap <Leader>JS <Leader>js
""""""""""""""""""""""""""""""""""""""""
```

> [!TIP]
>
> This uses the `!`_command_ feature to run an external command as a
> [filter][].
>
> Use the Vim command `:help mapleader` (or see [mapleader][]) for an
> explanation of `<Leader>`.


## Developing json-indent

See [DEVELOPING](DEVELOPING.md).


## References

- [Python][]:
    - [How to Run Your Python Scripts][run-python-scripts]
    - argcomplete on [PyPI][argcomplete-pypi] and [GitHub][argcomplete-github]
- [pre-commit][]
- [Vim][]:
    - [Intro to vimrc][vimrc]
    - [Vim filters][filter]
    - [Vim's Leader key][mapleader]
- [js-beautify][]


 [Python]: https://www.python.org/
 [run-python-scripts]: https://realpython.com/run-python-scripts/
 [uv]: https://github.com/astral-sh/uv

 [argcomplete-pypi]: https://pypi.org/project/argcomplete/
 [argcomplete-github]: https://github.com/kislyuk/argcomplete
 [autocompletion]: https://en.wikipedia.org/wiki/Autocomplete
 [Bash]: https://en.wikipedia.org/wiki/Bash_(Unix_shell)

 [pre-commit]: https://pre-commit.com/

 [Vim]: http://www.vim.org/
 [vimrc]: https://vimhelp.org/usr_05.txt.html#vimrc-intro
 [filter]: https://vimhelp.org/change.txt.html#filter
 [mapleader]: https://vimhelp.org/map.txt.html#mapleader

 [js-beautify]: https://github.com/beautify-web/js-beautify
